# -*- coding: utf-8 -*-

"""
Extract dataset from Arkindex using API.
"""

import logging
import os
import pathlib
import random
import uuid
from collections import defaultdict

import imageio.v2 as iio
from arkindex import ArkindexClient, options_from_env
from tqdm import tqdm

from dan import logger
from dan.datasets.extract.utils import (
    insert_token,
    parse_tokens,
    save_image,
    save_json,
    save_text,
)

IMAGES_DIR = "images"  # Subpath to the images directory.
LABELS_DIR = "labels"  # Subpath to the labels directory.
MANUAL_SOURCE = "manual"


def parse_worker_version(worker_version_id):
    if worker_version_id == MANUAL_SOURCE:
        return False
    return worker_version_id


def add_extract_parser(subcommands) -> None:
    parser = subcommands.add_parser(
        "extract",
        description=__doc__,
        help=__doc__,
    )

    # Required arguments.
    parser.add_argument(
        "--parent",
        type=uuid.UUID,
        nargs="+",
        help="ID of the parent folder to import from Arkindex.",
        required=False,
    )
    parser.add_argument(
        "--element-type",
        nargs="+",
        type=str,
        help="Type of elements to retrieve",
        required=True,
    )
    parser.add_argument(
        "--output",
        type=pathlib.Path,
        help="Path where the data will be generated.",
        required=True,
    )

    # Optional arguments.
    parser.add_argument(
        "--load-entities", action="store_true", help="Extract text with their entities"
    )
    parser.add_argument(
        "--tokens",
        type=pathlib.Path,
        help="Mapping between starting tokens and end tokens. Needed for entities.",
        required=False,
    )

    parser.add_argument(
        "--use-existing-split",
        action="store_true",
        help="Use the specified folder IDs for the dataset split.",
    )

    parser.add_argument(
        "--train-folder",
        type=uuid.UUID,
        help="ID of the training folder to import from Arkindex.",
        required=False,
    )
    parser.add_argument(
        "--val-folder",
        type=uuid.UUID,
        help="ID of the validation folder to import from Arkindex.",
        required=False,
    )
    parser.add_argument(
        "--test-folder",
        type=uuid.UUID,
        help="ID of the testing folder to import from Arkindex.",
        required=False,
    )

    parser.add_argument(
        "--transcription-worker-version",
        type=parse_worker_version,
        help=f"Filter transcriptions by worker_version. Use {MANUAL_SOURCE} for manual filtering.",
        required=False,
        default=MANUAL_SOURCE,
    )
    parser.add_argument(
        "--entity-worker-version",
        type=parse_worker_version,
        help=f"Filter transcriptions entities by worker_version. Use {MANUAL_SOURCE} for manual filtering.",
        required=False,
        default=MANUAL_SOURCE,
    )

    parser.add_argument(
        "--train-prob", type=float, default=0.7, help="Training set split size."
    )

    parser.add_argument(
        "--val-prob", type=float, default=0.15, help="Validation set split size"
    )

    parser.set_defaults(func=run)


class ArkindexExtractor:
    """
    Extract data from Arkindex
    """

    def __init__(
        self,
        client,
        folders,
        element_type,
        split_names,
        output,
        load_entities,
        tokens,
        use_existing_split,
        transcription_worker_version,
        entity_worker_version,
        train_prob,
        val_prob,
    ) -> None:
        self.client = client
        self.element_type = element_type
        self.split_names = split_names
        self.output = output
        self.load_entities = load_entities
        self.tokens = parse_tokens(tokens) if self.load_entities else None
        self.use_existing_split = use_existing_split
        self.transcription_worker_version = transcription_worker_version
        self.entity_worker_version = entity_worker_version
        self.train_prob = train_prob
        self.val_prob = val_prob

        self.get_subsets(folders)

    def get_subsets(self, folders):
        if self.use_existing_split:
            self.subsets = [
                (folder, split) for folder, split in zip(folders, self.split_names)
            ]
        else:
            self.subsets = [(folder, None) for folder in folders]

    def assign_random_split(self):
        """
        assuming train_prob + valid_prob + test_prob = 1
        """
        prob = random.random()
        if prob <= self.train_prob:
            return self.split_names[0]
        elif prob <= self.train_prob + self.val_prob:
            return self.split_names[1]
        else:
            return self.split_names[2]

    def extract_transcription(
        self,
        element,
    ):
        transcriptions = self.client.request(
            "ListTranscriptions",
            id=element["id"],
            worker_version=self.transcription_worker_version,
        )

        if transcriptions["count"] != 1:
            logger.warning(
                f"More than one transcription found on element ({element['id']}) with this config."
            )
            return

        transcription = transcriptions["results"].pop()
        if self.load_entities:
            entities = self.client.request(
                "ListTranscriptionEntities",
                id=transcription["id"],
                worker_version=self.entity_worker_version,
            )
            if entities["count"] == 0:
                logger.warning(
                    f"No entities found on transcription ({transcription['id']})."
                )
                return
            else:
                text = transcription["text"]

            count = 0
            for entity in entities["results"]:
                matching_tokens = self.tokens[entity["entity"]["metas"]["subtype"]]
                start_token, end_token = (
                    matching_tokens["start"],
                    matching_tokens["end"],
                )
                text, count = insert_token(
                    text,
                    count,
                    start_token,
                    end_token,
                    offset=entity["offset"],
                    length=entity["length"],
                )
        else:
            text = transcription["text"].strip()
        return text

    def process_element(
        self,
        element,
        split,
    ):
        text = self.extract_transcription(
            element,
        )

        if not text:
            logging.warning(
                f"Skipping {element['id']} (zero or multiple transcriptions with worker_version=None)"
            )
        else:
            logging.info(f"Processed {element['type']} {element['id']}")

            im_path = os.path.join(
                self.output, IMAGES_DIR, split, f"{element['type']}_{element['id']}.jpg"
            )
            txt_path = os.path.join(
                self.output, LABELS_DIR, split, f"{element['type']}_{element['id']}.txt"
            )

            save_text(txt_path, text)
            try:
                image = iio.imread(element["zone"]["url"])
                save_image(im_path, image)
            except Exception:
                logger.error(f"Couldn't retrieve image of element ({element['id']}")
                pass
            return element["id"]

    def process_page(
        self,
        page,
        split,
    ):
        # Extract only pages
        data = defaultdict(list)
        if self.element_type == ["page"]:
            data["page"] = [
                self.process_element(
                    page,
                    split,
                )
            ]
        # Extract page's children elements (text_zone, text_line)
        else:
            for element_type in self.element_type:
                for element in self.client.paginate(
                    "ListElementChildren",
                    id=page["id"],
                    type=element_type,
                    recursive=True,
                ):
                    data[element_type].append(
                        self.process_element(
                            element,
                            split,
                        )
                    )
        return data

    def run(self):
        split_dict = defaultdict(dict)
        # Iterate over the subsets to find the page images and labels.
        for subset_id, subset_split in self.subsets:
            page_idx = 0
            # Iterate over the pages to create splits at page level.
            for page in tqdm(
                self.client.paginate(
                    "ListElementChildren", id=subset_id, type="page", recursive=True
                )
            ):
                page_idx += 1
                split = subset_split or self.assign_random_split()

                split_dict[split][page["id"]] = self.process_page(
                    page=page,
                    split=split,
                )

        save_json(self.output / "split.json", split_dict)


def run(
    parent,
    element_type,
    output,
    load_entities,
    tokens,
    use_existing_split,
    train_folder,
    val_folder,
    test_folder,
    transcription_worker_version,
    entity_worker_version,
    train_prob,
    val_prob,
):
    assert (
        use_existing_split or parent
    ), "One of `--use-existing-split` and `--parent` must be set"

    if use_existing_split:
        assert (
            train_folder
        ), "If you use an existing split, you must specify the training folder."
        assert (
            val_folder
        ), "If you use an existing split, you must specify the validation folder."
        assert (
            test_folder
        ), "If you use an existing split, you must specify the testing folder."
        folders = [train_folder, val_folder, test_folder]
    else:
        folders = parent

    if load_entities:
        assert tokens, "Please provide the entities to match."

    # Login to arkindex.
    assert (
        "ARKINDEX_API_URL" in os.environ
    ), "The ARKINDEX API URL was not found in your environment."
    assert (
        "ARKINDEX_API_TOKEN" in os.environ
    ), "Your API credentials was not found in your environment."
    client = ArkindexClient(**options_from_env())

    # Create out directories
    split_names = ["train", "val", "test"]
    for split in split_names:
        os.makedirs(os.path.join(output, IMAGES_DIR, split), exist_ok=True)
        os.makedirs(os.path.join(output, LABELS_DIR, split), exist_ok=True)

    ArkindexExtractor(
        client=client,
        folders=folders,
        element_type=element_type,
        split_names=split_names,
        output=output,
        load_entities=load_entities,
        tokens=tokens,
        use_existing_split=use_existing_split,
        transcription_worker_version=transcription_worker_version,
        entity_worker_version=entity_worker_version,
        train_prob=train_prob,
        val_prob=val_prob,
    ).run()
